// Copyright 2016-2019 Doug Moen
// Licensed under the Apache License, version 2.0
// See accompanying file LICENSE or https://www.apache.org/licenses/LICENSE-2.0

//----------------------------------------------------------------------------
// The REPL is the Read-Eval-Print Loop. It's the interactive command line
// shell that lets you type Curv expressions, actions and definitions.
//
// It has two threads: a Repl thread that runs the REPL, and a Viewer thread
// that displays shapes in a Viewer window, and runs the OpenGL frame loop.
// The Viewer thread is a "worker" thread that receives messages from
// the REPL thread telling it to display a new shape or exit.
//
// On macOS, windows can only be run in the main thread, so the Viewer thread
// must be the main thread, and we must spawn a new thread to run the REPL.
// This is an inversion of the usual practice of running workers in spawned
// threads.

extern "C" {
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <unistd.h>
}
#include <iostream>
#include <fstream>
#include <thread>
#include <condition_variable>
#include <functional>

#include <replxx.hxx>

#include "shapes.h"
#include "view_server.h"

#include <libcurv/ansi_colour.h>
#include <libcurv/context.h>
#include <libcurv/exception.h>
#include <libcurv/program.h>
#include <libcurv/source.h>
#include <libcurv/system.h>

#include <libcurv/shape.h>
#include <libcurv/geom/viewer/viewer.h>

View_Server view_server;

bool was_interrupted = false;

void interrupt_handler(int)
{
    was_interrupted = true;
}

replxx::Replxx::completions_t get_completions(std::string const& context, int index, void* user_data)
{
    auto* names = static_cast<curv::Namespace*>(user_data);
    replxx::Replxx::completions_t completions;

    std::string prefix {context.substr(index)};
    for (auto const& n : *names) {
        if (n.first.size() < prefix.size())
            continue;

        if (prefix.compare(0, std::string::npos, n.first.c_str(), prefix.size()) == 0)
            completions.emplace_back(n.first.c_str());
    }

    return completions;
}

void color_input(std::string const& context, replxx::Replxx::colors_t& colors, void* user_data)
{
  auto* sys = static_cast<curv::System*>(user_data);

  auto source = curv::make<curv::String_Source>("", context);

  try {
    curv::Scanner scanner{std::move(source), *sys};

    for (;;) {
      using Color = replxx::Replxx::Color;
      using curv::Token;

      Color col;
      curv::Token tok = scanner.get_token();

      switch (tok.kind_) {
        case Token::k_end:
          return;

        // keywords
        case Token::k_by:
        case Token::k_do:
        case Token::k_else:
        case Token::k_for:
        case Token::k_if:
        case Token::k_in:
        case Token::k_let:
        case Token::k_include:
        case Token::k_var:
        case Token::k_where:
        case Token::k_while:
          col = Color::BRIGHTMAGENTA;
          break;

        // numerals
        case Token::k_num:
        case Token::k_hexnum:
          col = Color::BROWN;
          break;

        // string literal
        case Token::k_quote:
        case Token::k_string_segment:
        case Token::k_char_escape:
        case Token::k_backtick:
          col = Color::BRIGHTGREEN;
          break;

        // string interpolation
        case Token::k_dollar_paren:
        case Token::k_dollar_brace:
        case Token::k_dollar_bracket:
        case Token::k_dollar_ident:
          col = Color::BROWN;
          break;

        // syntactic symbols
        case Token::k_semicolon:
        case Token::k_equate:
        case Token::k_assign:
        case Token::k_ellipsis:
        case Token::k_at:
        case Token::k_right_arrow:
          col = Color::BRIGHTRED;
          break;

        // operators
        case Token::k_power:
        case Token::k_plus:
        case Token::k_minus:
        case Token::k_times:
        case Token::k_over:
        case Token::k_range:
        case Token::k_open_range:
        case Token::k_equal:
        case Token::k_not_equal:
        case Token::k_less:
        case Token::k_less_or_equal:
        case Token::k_greater:
        case Token::k_greater_or_equal:
        case Token::k_not:
        case Token::k_and:
        case Token::k_or:
        case Token::k_left_call:
        case Token::k_right_call:
          col = Color::BRIGHTBLUE;
          break;

        default:
          col = Color::DEFAULT;
          break;
      }

      auto start = colors.begin();
      std::fill(start + tok.first_, start + tok.last_, col);
    }
  } catch (curv::Exception&) {}
}

// REPL_Executor is used to execute a command line program that is a statement.
// Both push_value() and push_field() are supported.
// In most cases, the value or field is printed on the console.
// If there is 1 push_value and no push_field calls, then I test if the value
// is a shape, and display the shape graphically.
//
// Future: With multi-viewer support, 'cube;sphere' could open multiple Viewers.
// I don't know how reuse of already open viewer windows would work.
struct REPL_Executor : public curv::Operation::Executor
{
    curv::Namespace& names_;
    bool has_produced_output_ = false;
    curv::Value only_output_so_far_is_this_value_ = curv::missing;

    REPL_Executor(curv::Namespace& n) : names_(n) {}

    void start_command()
    {
        has_produced_output_ = false;
        only_output_so_far_is_this_value_ = curv::missing;
    }
    void push_value(curv::Value val, const curv::Context&) override
    {
        if (!has_produced_output_) {
            only_output_so_far_is_this_value_ = val;
            has_produced_output_ = true;
        } else {
            if (!only_output_so_far_is_this_value_.eq(curv::missing)) {
                std::cout << only_output_so_far_is_this_value_ << "\n";
                only_output_so_far_is_this_value_ = curv::missing;
            }
            std::cout << val << "\n";
        }
    }
    void push_field(curv::Symbol_Ref name, curv::Value val, const curv::Context&)
    override
    {
        if (!only_output_so_far_is_this_value_.eq(curv::missing)) {
            std::cout << only_output_so_far_is_this_value_ << "\n";
            only_output_so_far_is_this_value_ = curv::missing;
        }
        std::cout << name << ":" << val << "\n";
        has_produced_output_ = true;
    }
    void end_command(curv::Program& prog)
    {
        if (!only_output_so_far_is_this_value_.eq(curv::missing)) {
            curv::Value val = only_output_so_far_is_this_value_;
            static curv::Symbol_Ref lastval_key = "_";
            names_[lastval_key] = curv::make<curv::Builtin_Value>(val);
            curv::Shape_Program shape{prog};
            if (shape.recognize(val)) {
                print_shape(shape);
                view_server.display_shape(shape);
            } else {
                std::cout << val << "\n";
            }
        }
    }
};

void repl(curv::System* sys)
{
    // Catch keyboard interrupts, and set was_interrupted = true.
    // TODO: This will be used to interrupt the evaluator.
    if (isatty(0)) {
        struct sigaction interrupt_action;
        memset((void*)&interrupt_action, 0, sizeof(interrupt_action));
        interrupt_action.sa_handler = interrupt_handler;
        sigaction(SIGINT, &interrupt_action, nullptr);
    }

    // top level definitions, extended by typing 'id = expr'
    curv::Namespace names = sys->std_namespace();

    replxx::Replxx rx;
    rx.set_completion_callback(get_completions, static_cast<void*>(&names));
    rx.set_highlighter_callback(color_input, static_cast<void*>(sys));

    REPL_Executor executor{names};

    for (;;) {
        was_interrupted = false;
        const char* line = rx.input(AC_PROMPT "curv> " AC_RESET);
        if (line == nullptr) {
            if (errno == EAGAIN) continue;
            std::cout << "\n";
            break;
        }
        if (line[0] != '\0')
            rx.history_add(line);

        try {
            auto source = curv::make<curv::String_Source>("", line);
            curv::Program prog{std::move(source), *sys};
            prog.compile(&names);
            executor.start_command();
            auto bindings = prog.exec(executor);
            executor.end_command(prog);
            if (bindings) {
                for (auto f : *bindings)
                    names[f.first] = curv::make<curv::Builtin_Value>(f.second);
            }
        } catch (std::exception& e) {
            sys->error(e);
        }
    }
    view_server.exit();
}

void interactive_mode(
    curv::System& sys, const curv::geom::viewer::Viewer_Config& opts)
{
    sys.use_colour_ = true;
    std::thread repl_thread(repl, &sys);
    view_server.run(opts);
    if (repl_thread.joinable())
        repl_thread.join();
}
